////////////////////////////////////////////////////
// kello_hw4_v09.c
////////////////////////////////////////////////////
// Copyright 2017-2018 Jaakko Kairus
// jaakko@kair.us
//
// You may freely use and modify this software for personal use.
// Commercial use of this software or parts of it is prohibited
// without written permission from author.
// DISCLAIMER: This software comes with ABSOLUTELY NO WARRANTY
//
// For more information about this project see project web page at
// http://kair.us/projects/clock/v4/index.html
//
// Compiled with CCS compiler version 5.076
// Total compiled length 12436 words
///////////////////////////////////////////////////
// Version history
// v0.1 13.2.2017
//      Initial version. Support for TLC5925 and PCF2129. Compiler 5.054. Length 1260 words.
// v0.2 30.6.2017
//		Added initial support for remote (NEC coding). Length 1771 words.
// v0.3 6.9.2017
//		Moved remote decoding stuff to separate file (ir_machine.c) and changed remote
//		data type to a struct compatible with irmp (https://www.mikrocontroller.net/articles/IRMP)
//		Not however using IRMP to avoid getting involved with all the GNU GPL stuff.
//		Added support for showing remote control codes and learn mode to teach clock
//		to use any NEC protocol remote. Learned codes are stored to PIC EEPROM. Added
//		support to set date and time. Added support to adjust brightness. Brightness
//		and display mode are stored to PCF2129 NVRAM (using alarm registers).
//		Added temperature readout for two 10k NTC sensors. Fixed updating the date
//		when day changes. Length 7145 words.
// v0.4 8.10.2017
//		Changed ir_machine polling interval from 100s to 64s. This allows using any
//		of the timers to generate the required polling cycle. Changed timer interrupts
//		from timer 2 to timer 0 and 64 s. Frees timer 2 for driving buzzer.
//		Added hour beep function with possibility to set to short, long or off.
//		Better joystick operation with debounced press and release. Length 7457 words.
// v0.5 6.12.2017
//		Added support for RC5, Samsung and JVC IR protocols. Protocol is recognized 
//		automatically by the ir_machine() function. The identify menu now also shows
//		these codings. Length 7634 words.
// v0.6 23.12.2017
//		Added support for SIRC and NEC42 IR protocols. Rewrote RC5 decoder to sync
//		on every bit. Works now more reliably with different RC5 remotes. Added
//		hrbeep saving. Fixed displaying negative temperature values. Changed brightness
//		to show milliamperes instead of DAC values. Length 8503 words.
// v0.7 26.12.2017
//		Added support for time zones and automatic / manual daylight saving time.
//		Changed time registers to user struct_tm from time.h. Two structs, RTC uses always
//		utc_time. When performing RTC read, conversion from utc_time to local_time
//		is done, taking into account time zone and daylight saving time.
//		New include file timezone_converter.c has functions to deal with time zones
//		and daylight saving time. Added functions to pcf2129.c which work with struct_tm.
//		RTC is now read at 1 hour period instead of old 1 day period. This allows
//		proper way of checking daylight saving time. This approach needed the set time
//		and set date menus to update the RTC after each adjustment to prevent possible
//		time jump if minutes roll over to 0, causing RTC read of the 'old' time and date.
//		This causes some blinking in dimmed digits because the RTC read takes some 7 ms.
//		
//		9141 words when gps preliminary added, time in separate variables
//		9285 words when time changed to use struct_tm
//		9404 words when get_time_and_date changed to use struct_tm
//		9590 words when set_time_and_date changed to use struct_tm
//		11823 words when timezone and daylight saving time are in use
//		14007 words when GPS decoding active
//		13020 words with new GPRMC_decode_tm function which decodes time directly to struct_tm
//		bosch bme280 driver doesn't fit to program memory anymore. A total of about 17500 words
//		of program memory would be needed
//		12832 words with '#opt compress' optimization level defined and with bme280 driver,
//		support for showing humidity. Added decimal point location parameter to tlc5926
//		show number -function. Added support for detection of RTC and GPS at boot, and working
//		accordingly. Should work with any combination of these assembled, to be tested!
// v0.8 5.1.2018
//		Weird behavior with small unrelated changes in code cause e.g. temperature
//		displays to crash the FW. Updated compiler to 5.076. Program memory usage went
//		up from 12800 words to 15000 words, but the weird bugs disappeared. Now also
//		the bme280 seems to work properly in 'normal' mode, each bme280_get_humidity() is
//		returning fresh values without the need to first call bme280_set_mode() each time.
//		The #separate is needed again in bme280.c in order for the get_humidity function
//		to fit. Added support for displaying also bme280 temperature and pressure. bme280
//		is recognized at startup. The menu items to display bme280 sensor data are hidden
//		if the sensor is not found. Length 14757 words.
// v0.9 3.2.2018
//		Added option to show_number function to show temperature in same format as
//		show_temperature function. However using show_number function instead of
//		show_temperature functions consumes more program memory. Using new option
//		only on displaying BME temperatures of < -9.9C. Reformatted menu_machine()
//		function to use adjust_with_limits() function to scroll menu items. This
//		effectively avoids repeating event readout and state selection code.
//		The implementation required some reorganization of states and also usage
//		of some redirecting states.
//		14427 words after above changes
//		14267 words after changing tlc5925_show_string() to use 2-dimensional ascii array
//					and show_number and temperature functions to use 2-dimensional segment array
//		14639 words after changing tlc5925 to use more complete, 64 char ascii table
//		14644 words before changing tlc5925 to use display buffer for writes
//		14474 words just the unused 6-byte buffer added
//		14506 words when using tlc5925_send_buffer on show_temperature and show_number functions
//		14440 words when tlc5925_show_time() using get_digits() and tlc5925_send_buffer()
//		14408 words when tlc5925_show_date() using get_digits() and tlc5925_send_buffer()
//		14377 words when tlc5925_show_hex() using get_digits() and tlc5925_send_buffer()
//		14290 words when tlc5925_show_time_with_dimmer() using get_digits() and tlc5925_send_buffer()
//		14069 words when tlc5925_show_date_with_dimmer() using get_digits() and tlc5925_send_buffer()
//		now also fits to flash without #opt compress! 15485 words.
//		14235 words when bme_temperature shown on either in or out temperature if either ntc is not present
//		15522 words w/o opt compress.
//		12532 words when in and out temperatures use tlc5925_show_number function instead of show_temperature
//		15328 words w/o #opt compress.
//		14069 words when fixed seconds jumping with both RTC and GPS present. Required checking hour beep on
//		both RTC ISR and GPS_decode functions. using separate hourBeep() funtion for checking
//		If hourBeep() from GPS_decode commented out, memory usage 12514(!).
//		15338 words w/o #opt compress
//		12436 words when rewrote leading zero blanking and minus sign add routines to tlc5925_show_number funtion
//		14940 words w/o #opt compress


#opt compress
#define FW_VERSION "v0.9 "
#define FW_VERSION 90
#define OPAMP_NI_TO_DAC1 0x04	// Needed because error in CCS include file

#define RTC_CLK PIN_A6
#define TLC5925_LE PIN_C0

#define IRin PIN_A5
#define IRED PIN_A7

#define REXT 1000	// TLC5925 current set resistor value in ohms

#include <16F1788.h>
#device ADC=12;
#DEVICE PASS_STRINGS=IN_RAM // allows passing constant char arrays to tlc5925.c function
							// copies string from rom to ram when needed
#include <stdlib.h>
#include <stdio.h>
#define TIME_T_USES_2010
#include <time.h>
#include <internal_eeprom.c>	// for reading / writing int16
#fuses INTRC_IO,NOWDT,NOFCMEN,NOIESO,BROWNOUT,BORV25,NOPROTECT,NOLVP,NODEBUG,PUT,NOMCLR,NOCLKOUT,NOVCAP
#use delay(int=16MHz, clock=32MHz)
#use FAST_IO(a)
#use FAST_IO(b)
#use FAST_IO(c)
#use FAST_IO(e)
#use spi(FORCE_HW,BAUD=1000000, BITS=8, SAMPLE_RISE)
#use i2c(master,sda=PIN_B3,scl=PIN_B0,fast=50000,stream=softai2c)
#define STREAM_I2C_BME280 softai2c
#use pwm(CCP1, OUTPUT=PIN_C2, TIMER=2, PWM_OFF)
#use rs232(baud=115200,xmit=PIN_C6,rcv=PIN_C7,ERRORS)
#include "pcf2129.c"
#include "tlc5925.c"
#include "ir_machine.c"
#include "ntc_temperature.c"
#include "gprmc_decoder.c"
#include "timezone_converter.c"
#include "bme280.c"

#define REMOTE_PROTOCOL_ADDR 0
#define REMOTE_ADDRESS_ADDR 2
#define REMOTE_UP_ADDR 4
#define REMOTE_DOWN_ADDR 6
#define REMOTE_LEFT_ADDR 8
#define REMOTE_RIGHT_ADDR 10
#define REMOTE_BUTTON_ADDR 12
#define TIMEZONE_ADDR 14
#define DLIGHT_SAVING_ADDR 15
#define BRIGHTNESS_ADDR 16
#define HRBEEP_ADDR 17

typedef enum {
        ST_MENU_SET = 0,
        ST_MENU_TIME,
        ST_MENU_DATE,
        ST_MENU_IN_TEMPERATURE,
        ST_MENU_OUT_TEMPERATURE,
        ST_MENU_TOGGLE,
        ST_MENU_HUMIDITY,
        ST_MENU_PRESSURE,
        ST_MENU_SET_REDIRECT,
        ST_TIME,
        ST_DATE,
        ST_IN_TEMPERATURE,
        ST_OUT_TEMPERATURE,
        ST_TOGGLE,
        ST_HUMIDITY,
        ST_PRESSURE,
        ST_MENU_SET_TIME,
        ST_MENU_SET_DATE,
        ST_MENU_SET_ZONE,
        ST_MENU_SET_SAVING,
        ST_MENU_SET_BRIGHTNESS,
        ST_MENU_SET_HRBEEP,
        ST_MENU_SET_REMOTE,
        ST_MENU_SET_VERSION,
        ST_MENU_SET_RETURN,
        ST_SET_HOURS,
        ST_SET_DAY,
        ST_SET_ZONE,
        ST_SET_SAVING,
        ST_SET_BRIGHTNESS,
        ST_SET_HRBEEP,
        ST_MENU_REMOTE_IDENTIFY_REDIRECT,
        ST_VERSION,
        ST_MENU_SET_RETURN_REDIRECT,
        ST_SET_MINUTES,
        ST_SET_SECONDS,
        ST_SET_MONTH,
        ST_SET_YEAR,
        ST_MENU_REMOTE_IDENTIFY,
        ST_MENU_REMOTE_SHOW_CODE,
        ST_MENU_REMOTE_LEARN,
        ST_MENU_REMOTE_RETURN,
        ST_REMOTE_IDENTIFY,
        ST_REMOTE_SHOW_CODE,
        ST_LEARN_REMOTE_UP,
        ST_MENU_REMOTE_RETURN_REDIRECT,
        ST_LEARN_REMOTE_DOWN,
        ST_LEARN_REMOTE_LEFT,
        ST_LEARN_REMOTE_RIGHT,
        ST_LEARN_REMOTE_BUTTON,
        ST_UNDEFINED
} MENU_STATES;
MENU_STATES menu_state = ST_TIME;

typedef enum {
        EVT_NONE = 0,
        EVT_UP,
        EVT_DOWN,
        EVT_RIGHT,
        EVT_LEFT,
        EVT_BUTTON
//        EVT_LONG_PRESS
} EVENTS;
EVENTS event = EVT_NONE;


int8 brightness=6;
int8 hrbeep=0;
int16 buzzerDuration=0;
int8 dlight_saving=0;
int8 dlight_saving_on=0;
signed int8 timezone=0;

struct_tm local_time;

unsigned int8 joyshadow, joytimer,releasetimer;

int1 ir_received = FALSE;

// GPS stuff:
char GPSData[128];
unsigned int8 GPSDataPtr=0;
char c;
int8 GPSDataReady = FALSE;
int8 GPS_status = 2;	// 0=no fix, 1=fix, 2=no GPS module found

int8 RTC_init_status;	// 0=OK, 1=osc stopped, 2=no RTC found
int8 bme_ok;			// 0=not initialized / found, 1=OK

int8 pending=0;
#bit PENDING_RTC_SAVE = pending.0
#bit PENDING_RTC_READ = pending.1
#bit PENDING_STATE_SAVE = pending.2
#bit PENDING_BRIGHTNESS_SAVE = pending.3
#bit PENDING_REMOTE_SAVE = pending.4
#bit PENDING_HRBEEP_SAVE = pending.5
#bit PENDING_BRIGHTNESS_SET = pending.6
#bit PENDING_SETTINGS_SAVE = pending.7

IR_PACKET receivedPacket;

int8 remoteProtocol;
int16 remoteAddress;
int16 remoteCommandUp;
int16 remoteCommandDown;
int16 remoteCommandLeft;
int16 remoteCommandRight;
int16 remoteCommandButton;

int8 decode_allowed=1;

signed int32 bme280_temp=0;
unsigned int32 bme280_pressure=0;
unsigned int32 bme280_humidity=0;
bme280_init_t bme280_config;

struct_tm utc_time;
struct_tm gps_time;

void read_jstk(void);
void code_to_button(void);
void menuMachine();
void displayMachine();
void adjust_with_limits(signed int8* adjust, signed int8 low_limit, signed int8 high_limit);
void adjust_with_limits(int16* adjust, int16 low_limit, int16 high_limit);
void save_remote();
void load_remote();
void handle_pending_tasks();
void handleBuzzer();
void decodeGPS();
void my_write_eeprom(int8 address, int8 value);
int8 my_read_eeprom(int8 address);
void show_bme_temperature();
void hourBeep();

#int_ioc
void rtcclkisr(void) {
	if (GPS_status!=1) {	// only increment time based on RTC interrupts if no GPS found or no GPS fix
		if (local_time.tm_sec<59)
			local_time.tm_sec++;
		else {
			local_time.tm_sec=0;
			if (local_time.tm_min<59)
				local_time.tm_min++;
			else {
				local_time.tm_min=0;
				PENDING_RTC_READ=1;			// fetch new time from RTC each hour. this updates also dst when necessary
			}	
		}
		hourBeep();
	}	
}

#int_timer0
void timer0isr(void) {
	//output_high(IRED);	// for measuring interrupt length. Currently min 12 s when showing time. Typically max 20 s when decoding and comparing remote codes. Abs. max 30 s.
							// with #opt compress typically 19 s when showing time, does not vary too much when receiving codes except when complete code has been received
	ir_machine();
	ir_received = getIrPacket(&receivedPacket);

	if (ir_received && decode_allowed)
		code_to_button();

	read_ntc_sensors();

	read_jstk();
	menuMachine();
	if (buzzerDuration)
		buzzerDuration--;
	//output_low(IRED);
}	

#int_rda
void uartisr(void)		// uart interrupt for receiving GPS NMEA messages
{
   c = getc();
   if (GPSDataReady)
   	return;
   switch (c)
   {
      case '$':
         GPSDataPtr = 0;
      break;
     
      case '\n':
      	if (GPSData[3] == 'R')		// $GPRMC message
         GPSDataReady = TRUE;
      break;
   }
   
   GPSData[GPSDataPtr++ & 0x7F] = c;
     
} 

void main()
{
	unsigned int8 bme_sampling_delay=0;
	setup_timer_0(T0_DIV_2); // set timer0 to 64 s (prescaler 2)
	enable_interrupts(INT_TIMER0);
	enable_interrupts(INT_IOC_A6_L2H);	// interrupt for 1 Hz clock signal from RTC chip Changed from H2L to L2H, maybe fixes second jump bug (is the RTC updated at low-to-high -transition?
	enable_interrupts(INT_RDA);	// for receiving GPS meassages via uart
	enable_interrupts(GLOBAL);
	set_tris_a(0b01111111);		// all input except ired
	set_tris_b(0b11111111);		// tri-state all port b pins, scl and sda will be controlled by #use_i2c code
	set_tris_c(0b10010000);		// SDI and RX are inputs
	set_tris_e(0x11111000);		
	port_a_pullups(0b01111111);	// pullups for joystick, ir receiver and rtc_clk. not for ired
	port_b_pullups(0b11000000); // pullups for pgc and pgd lines
	port_c_pullups(0b10000000); // pullup for rx input incase no gps module installed
	port_e_pullups(0b00001000); // mclr pullup
		
	setup_adc_ports( sAN11|sAN13,VSS_VDD);	// Temperature sensors
	setup_vref(VREF_ON | VREF_ADC_2v048 | VREF_COMP_DAC_2v048);
	setup_opamp2(OPAMP_ENABLED | OPAMP_HIGH_GBWP_MODE | OPAMP_NI_TO_DAC1);
	setup_dac(DAC_VSS_FVR);

	setup_adc(ADC_CLOCK_DIV_32);	// Gives 1s conversion clock cycle time (fastest recommended)
	set_adc_channel(11);
	sums_ready=0;	// enables ADC conversions

	output_low(PIN_C2);	// ensure the off state of buzzer doesn't present a DC voltage over the buzzer

	setup_timer_2(T2_DIV_BY_64,125,1); // 1 kHz (8 MHz / 64 / 125)
	set_pwm1_duty(62);
	delay_ms(250);
	setup_ccp1(CCP_OFF);
	output_low(PIN_C1);	// /OE low = display on
	
	RTC_init_status = pcf2129_init();
	
	if (RTC_init_status==1) {
		pcf2129_set_byte(menu_state, PCF2129_SEC_ALARM_REG);	// Clock has been stopped because no battery installed
		pcf2129_set_byte(brightness, PCF2129_MIN_ALARM_REG);	// or first time powered
		pcf2129_set_byte(hrbeep, PCF2129_HOUR_ALARM_REG);
	}
	else if (RTC_init_status==0){
		pcf2129_get_byte(&menu_state, PCF2129_SEC_ALARM_REG);	// Alaram registers are used to store variables
		pcf2129_get_byte(&brightness, PCF2129_MIN_ALARM_REG);	// which change often
		pcf2129_get_byte(&hrbeep, PCF2129_HOUR_ALARM_REG);
	}
	else {														// No RTC chip found, read brightness and hrbeep settings from EEPROM
		brightness = read_eeprom(BRIGHTNESS_ADDR);
		hrbeep = read_eeprom(HRBEEP_ADDR);
	}	
		

	if (menu_state >= ST_UNDEFINED)		// Some sanity checks for values stored to SRAM
		menu_state = ST_TIME;
	if (brightness > 30)
		brightness = 30;

	timezone = read_eeprom(TIMEZONE_ADDR);
	dlight_saving = read_eeprom(DLIGHT_SAVING_ADDR);
	
	if (timezone < -12 || timezone > 12)
		timezone=0;
	if (dlight_saving>2)
		dlight_saving=0;
		
	load_remote();

	PENDING_RTC_READ = 1;
	PENDING_BRIGHTNESS_SET = 1;
	
	bme280_config.humidityOversampling = BM280_OSRS_16X;
	bme280_config.temperatureOversampling = BM280_OSRS_16X;
	bme280_config.pressureOversampling = BM280_OSRS_16X;
	bme280_config.inactiveDuration = BM280_TIMING_10;
	bme280_config.iirFilter = BM280_FILTER_OFF;
	bme280_init(bme280_config);
	bme_ok = bme280_set_mode(BM280_MODE_NORMAL);	// bme280_set_mode returns false if bme not initialized or found

	// Infinite loop
	while(1)
	{
		if (pending)
			handle_pending_tasks();

		displayMachine();


		if (menu_state <= ST_PRESSURE) {	// BME sensor is not read when in setupb
			bme_sampling_delay++;	// no zeroing, just rolls over and sampling happens then
			if (bme_sampling_delay==0) {
				bme280_get_humidity(&bme280_temp, &bme280_pressure, &bme280_humidity);
			}	
		}		

		decodeGPS();
		
		handleBuzzer();

		if (menu_state == ST_IN_TEMPERATURE || menu_state == ST_OUT_TEMPERATURE || menu_state == ST_TOGGLE) {
			lookup_32bit(&in_sum, &in_temp, &inhyst);
			lookup_32bit(&out_sum, &out_temp, &outhyst);
			sums_ready=0;
		}	

	}	// end while
}

void menuMachine()
{
	switch(menu_state)
	{
		case ST_MENU_SET:
		case ST_MENU_TIME:
		case ST_MENU_DATE:
		case ST_MENU_IN_TEMPERATURE:
		case ST_MENU_OUT_TEMPERATURE:
		case ST_MENU_TOGGLE:
		case ST_MENU_HUMIDITY:
		case ST_MENU_PRESSURE:
			switch(event)
			{
				case EVT_BUTTON:
					menu_state += ST_TIME - ST_MENU_TIME;
					PENDING_STATE_SAVE=1;
					break;
				case EVT_LEFT:
				case EVT_RIGHT:
					if (bme_ok)
						adjust_with_limits(&menu_state, ST_MENU_SET, ST_MENU_PRESSURE);
					else
						adjust_with_limits(&menu_state, ST_MENU_SET, ST_MENU_TOGGLE);
					break;
				default:
			}
			break;	
		case ST_MENU_SET_REDIRECT:
			menu_state = ST_MENU_SET_TIME;
			break;
		case ST_TIME:
		case ST_DATE:
		case ST_IN_TEMPERATURE:
		case ST_OUT_TEMPERATURE:
		case ST_TOGGLE:
		case ST_HUMIDITY:
		case ST_PRESSURE:
			switch(event)
			{
				case EVT_BUTTON:
					menu_state -= ST_TIME - ST_MENU_TIME;
					break;
			}
			break;	
		case ST_MENU_SET_TIME:
		case ST_MENU_SET_DATE:
		case ST_MENU_SET_ZONE:
		case ST_MENU_SET_SAVING:
		case ST_MENU_SET_BRIGHTNESS:
		case ST_MENU_SET_HRBEEP:
		case ST_MENU_SET_REMOTE:
		case ST_MENU_SET_VERSION:
		case ST_MENU_SET_RETURN:
			switch(event)
			{
				case EVT_BUTTON:
					menu_state += ST_SET_HOURS - ST_MENU_SET_TIME;
					break;
				case EVT_LEFT:
				case EVT_RIGHT:
					adjust_with_limits(&menu_state, ST_MENU_SET_TIME, ST_MENU_SET_RETURN);
					break;
				default:
			}
			break;
		case ST_SET_HOURS:
			switch(event)
			{
				case EVT_UP:
				case EVT_DOWN:
					adjust_with_limits(&local_time.tm_hour, 0, 23);
					PENDING_RTC_SAVE = 1;		// Added 2.1.2018 at v0.7 to fix hours jumping if minutes roll over, causing RTC read of old time
					break;
				case EVT_LEFT:
					menu_state = ST_SET_SECONDS;
					break;
				case EVT_RIGHT:
					menu_state = ST_SET_MINUTES;
					break;
				case EVT_BUTTON:
					menu_state = ST_MENU_SET_TIME;
					break;
			}
			break;
		case ST_SET_MINUTES:
			switch(event)
			{
				case EVT_UP:
				case EVT_DOWN:
					adjust_with_limits(&local_time.tm_min, 0, 59);
					PENDING_RTC_SAVE = 1;		// Added 2.1.2018 at v0.7 to fix minutes jumping if set to 59 and seconds roll over, causing RTC read of old time
					break;
				case EVT_BUTTON:
					menu_state = ST_MENU_SET_TIME;
					break;
				case EVT_LEFT:
					menu_state = ST_SET_HOURS;
					break;
				case EVT_RIGHT:
					menu_state = ST_SET_SECONDS;
					break;
			}
			break;
		case ST_SET_SECONDS:
			switch(event)
			{
				case EVT_BUTTON:
					menu_state = ST_MENU_SET_TIME;
					//PENDING_RTC_SAVE = 1;
					break;
				case EVT_LEFT:
					menu_state = ST_SET_MINUTES;
					break;
				case EVT_RIGHT:
					menu_state = ST_SET_HOURS;
					break;
				case EVT_NONE:
					break;
				default:
				//case EVT_UP:
				//case EVT_DOWN:
					if (local_time.tm_sec>29) {
						if (local_time.tm_min<59)
							local_time.tm_min++;
						else
							local_time.tm_min=0;
					}
					local_time.tm_sec=0;	
					PENDING_RTC_SAVE = 1;		// Added 2.1.2018 at v0.7
					break;
			}
			break;
		case ST_SET_DAY:
			switch(event)
			{
				case EVT_BUTTON:
					menu_state = ST_MENU_SET_DATE;
					break;
				case EVT_LEFT:
					menu_state = ST_SET_YEAR;
					break;
				case EVT_RIGHT:
					menu_state = ST_SET_MONTH;
					break;
				case EVT_NONE:
					break;
				default:
					adjust_with_limits(&local_time.tm_mday, 1, 31);
					PENDING_RTC_SAVE = 1;		// Added 2.1.2018 at v0.7
					break;
			}
			break;
		case ST_SET_MONTH:
			switch(event)
			{
				case EVT_BUTTON:
					menu_state = ST_MENU_SET_DATE;
					break;
				case EVT_LEFT:
					menu_state = ST_SET_DAY;
					break;
				case EVT_RIGHT:
					menu_state = ST_SET_YEAR;
					break;
				case EVT_NONE:
					break;
				default:
					adjust_with_limits(&local_time.tm_mon, 0, 11);
					PENDING_RTC_SAVE = 1;		// Added 2.1.2018 at v0.7
					break;
			}
			break;
		case ST_SET_YEAR:
			switch(event)
			{
				case EVT_BUTTON:
					menu_state = ST_MENU_SET_DATE;
					break;
				case EVT_LEFT:
					menu_state = ST_SET_MONTH;
					break;
				case EVT_RIGHT:
					menu_state = ST_SET_DAY;
					break;
				case EVT_NONE:
					break;
				default:
					adjust_with_limits(&local_time.tm_year, 118, 199);
					PENDING_RTC_SAVE = 1;		// Added 2.1.2018 at v0.7
					break;
			}
			break;
		case ST_SET_ZONE:
			switch(event)
			{
				case EVT_BUTTON:
					menu_state = ST_MENU_SET_ZONE;
					PENDING_SETTINGS_SAVE = 1;
					PENDING_RTC_READ = 1;
					break;
				case EVT_NONE:
					break;
				case EVT_UP:
				case EVT_DOWN:
					adjust_with_limits(&timezone, -12, 12);
					break;
				default:
			}
			break;
		case ST_SET_SAVING:
			switch(event)
			{
				case EVT_BUTTON:
					menu_state = ST_MENU_SET_SAVING;
					PENDING_SETTINGS_SAVE = 1;
					PENDING_RTC_READ = 1;
					break;
				case EVT_NONE:
					break;
				//default:
				case EVT_UP:
				case EVT_DOWN:
					adjust_with_limits(&dlight_saving, 0, 2);	// OFF, ON, AUTO
					break;
				default:
			}
			break;
		case ST_SET_BRIGHTNESS:
			switch(event)
			{
				case EVT_BUTTON:
					menu_state = ST_MENU_SET_BRIGHTNESS;
					PENDING_BRIGHTNESS_SAVE = 1;
					break;
				case EVT_NONE:
					break;
				//default:
				case EVT_UP:
				case EVT_DOWN:
					adjust_with_limits(&brightness, 3, 30);	// 28 current settings from 1.5 mA to 15 mA
					PENDING_BRIGHTNESS_SET = 1;
					break;
				default:
			}
			break;
		case ST_SET_HRBEEP:
			switch(event)
			{
				case EVT_BUTTON:
					menu_state = ST_MENU_SET_HRBEEP;
					PENDING_HRBEEP_SAVE = 1;
					break;
				case EVT_NONE:
					break;
				case EVT_UP:
				case EVT_DOWN:
					adjust_with_limits(&hrbeep, 0, 2);	// 0=none, 1=short, 2=long
					break;
				default:
			}
			break;
		case ST_MENU_REMOTE_IDENTIFY_REDIRECT:
			menu_state = ST_MENU_REMOTE_IDENTIFY;
			break;
		case ST_VERSION:
			switch(event)
			{
				case EVT_BUTTON:
					menu_state = ST_MENU_SET_VERSION;
					break;
			}
			break;
		case ST_MENU_SET_RETURN_REDIRECT:	// this pseudo state is just to allow constant
			menu_state = ST_MENU_SET;		// jumps from setup menu items
			break;
		case ST_MENU_REMOTE_IDENTIFY:
		case ST_MENU_REMOTE_SHOW_CODE:
		case ST_MENU_REMOTE_LEARN:
		case ST_MENU_REMOTE_RETURN:
			decode_allowed=1;
			switch(event)
			{
				case EVT_BUTTON:
					menu_state += ST_REMOTE_IDENTIFY - ST_MENU_REMOTE_IDENTIFY;
					break;
				case EVT_LEFT:
				case EVT_RIGHT:
					adjust_with_limits(&menu_state, ST_MENU_REMOTE_IDENTIFY, ST_MENU_REMOTE_RETURN); 
					break;
			}
			break;
		case ST_REMOTE_IDENTIFY:
			switch(event)
			{
				case EVT_BUTTON:
					menu_state = ST_MENU_REMOTE_IDENTIFY;
					break;
			}
			break;
		case ST_REMOTE_SHOW_CODE:
			switch(event)
			{
				case EVT_BUTTON:
					menu_state = ST_MENU_REMOTE_SHOW_CODE;
					break;
			}
			break;
		case ST_LEARN_REMOTE_UP:
			decode_allowed=0;
			switch(event)
			{
				case EVT_BUTTON:
					menu_state = ST_MENU_REMOTE_LEARN;
					break;
			}
			if (ir_received) {
				remoteProtocol = receivedPacket.protocol;
				remoteAddress = receivedPacket.address;
				remoteCommandUp = receivedPacket.command;
				ir_received = FALSE;
				menu_state = ST_LEARN_REMOTE_DOWN;
			}	
			break;
		case ST_MENU_REMOTE_RETURN_REDIRECT:
			menu_state = ST_MENU_SET_REMOTE;
			break;
		case ST_LEARN_REMOTE_DOWN:
			switch(event)
			{
				case EVT_BUTTON:
					menu_state = ST_MENU_REMOTE_LEARN;
					break;
			}
			if (ir_received) {
				remoteCommandDown = receivedPacket.command;
				ir_received = FALSE;
				menu_state = ST_LEARN_REMOTE_LEFT;
			}	
			break;
		case ST_LEARN_REMOTE_LEFT:
			switch(event)
			{
				case EVT_BUTTON:
					menu_state = ST_MENU_REMOTE_LEARN;
					break;
			}
			if (ir_received) {
				remoteCommandLeft = receivedPacket.command;
				ir_received = FALSE;
				menu_state = ST_LEARN_REMOTE_RIGHT;
			}	
			break;
		case ST_LEARN_REMOTE_RIGHT:
			switch(event)
			{
				case EVT_BUTTON:
					menu_state = ST_MENU_REMOTE_LEARN;
					break;
			}
			if (ir_received) {
				remoteCommandRight = receivedPacket.command;
				ir_received = FALSE;
				menu_state = ST_LEARN_REMOTE_BUTTON;
			}	
			break;
		case ST_LEARN_REMOTE_BUTTON:
			switch(event)
			{
				case EVT_BUTTON:
					menu_state = ST_MENU_REMOTE_LEARN;
					break;
			}
			if (ir_received) {
				remoteCommandButton = receivedPacket.command;
				ir_received = FALSE;
				menu_state = ST_MENU_REMOTE_LEARN;
				PENDING_REMOTE_SAVE = 1;
			}	
			break;
	} // end menu case
	event = EVT_NONE;
}

void displayMachine()
{
	switch(menu_state)
	{
		case ST_MENU_TIME:
			tlc5925_show_string("hour");
			break;	
		case ST_MENU_DATE:
			tlc5925_show_string("date");
			break;	
		case ST_MENU_IN_TEMPERATURE:
			tlc5925_show_string("intemp");
			break;	
		case ST_MENU_OUT_TEMPERATURE:
			tlc5925_show_string("outemp");
			break;
		case ST_MENU_HUMIDITY:
			tlc5925_show_string("humidi");
			break;
		case ST_MENU_PRESSURE:
			tlc5925_show_string("pressu");
			break;
		case ST_MENU_TOGGLE:
			tlc5925_show_string("toggle");
			break;
		case ST_MENU_SET:
			tlc5925_show_string("setup");
			break;
		case ST_TIME:
			tlc5925_show_time(local_time.tm_hour, local_time.tm_min, local_time.tm_sec);
			break;
		case ST_DATE:
			tlc5925_show_date(local_time.tm_mday, local_time.tm_mon+1, (int8)(local_time.tm_year));
			break;
		case ST_IN_TEMPERATURE:
			if (in_temp != -400)
				//tlc5925_show_temperature(in_temp);
				tlc5925_show_number((signed int32)(in_temp*10),(2|0x20));
				//tlc5925_show_number(in_temp,(2|0x20));
			else if (bme_ok)
				show_bme_temperature();
			else
				tlc5925_show_string("n/c");
			break;
		case ST_OUT_TEMPERATURE:
			if (out_temp != -400)
				//tlc5925_show_temperature(out_temp);
				tlc5925_show_number((signed int32)(out_temp*10),(2|0x20));
				//tlc5925_show_number(out_temp,(2|0x20));
			else if (in_temp != -400 && bme_ok)
				show_bme_temperature();
			else
				tlc5925_show_string("n/c");
			break;
		case ST_HUMIDITY:
			tlc5925_show_number(bme280_humidity*100/1024, 2|0x08);
			//tlc5925_show_number((signed int32)(bme280_humidity*100/1024), 2|0x08);
			break;
		case ST_PRESSURE:
			tlc5925_show_number(bme280_pressure, 2);
			//tlc5925_show_number((signed int32)(bme280_pressure), 2);
			break;
		case ST_TOGGLE:
			if(local_time.tm_sec%10==0 || local_time.tm_sec%10==1)
				tlc5925_show_date(local_time.tm_mday, local_time.tm_mon+1, (int8)(local_time.tm_year));
			else
				tlc5925_show_time(local_time.tm_hour, local_time.tm_min, local_time.tm_sec);
			break;
		case ST_MENU_SET_TIME:
			tlc5925_show_string("sethou");
			break;	
		case ST_SET_HOURS:
			tlc5925_show_time_with_dimmer(local_time.tm_hour, local_time.tm_min, local_time.tm_sec, 1);	
			break;
		case ST_SET_MINUTES:
			tlc5925_show_time_with_dimmer(local_time.tm_hour, local_time.tm_min, local_time.tm_sec, 2);	
			break;
		case ST_SET_SECONDS:
			tlc5925_show_time_with_dimmer(local_time.tm_hour, local_time.tm_min, local_time.tm_sec, 3);	
			break;
		case ST_MENU_SET_DATE:
			tlc5925_show_string("setdat");
			break;	
		case ST_SET_DAY:
			tlc5925_show_date_with_dimmer(local_time.tm_mday, local_time.tm_mon+1, (int8)(local_time.tm_year), 1);
			break;
		case ST_SET_MONTH:
			tlc5925_show_date_with_dimmer(local_time.tm_mday, local_time.tm_mon+1, (int8)(local_time.tm_year), 2);
			break;
		case ST_SET_YEAR:
			tlc5925_show_date_with_dimmer(local_time.tm_mday, local_time.tm_mon+1, (int8)(local_time.tm_year), 3);
			break;
		case ST_MENU_SET_ZONE:
			tlc5925_show_string("t-zone");
			break;
		case ST_SET_ZONE:
			tlc5925_show_number(timezone,0x07);
			//tlc5925_show_number((signed int32)(timezone),0x07);
			break;
		case ST_MENU_SET_SAVING:
			tlc5925_show_string("daylig");
			break;
		case ST_SET_SAVING:
			switch (dlight_saving)
			{
				case 0:
					tlc5925_show_string("OFF");
					break;
				case 1:
					tlc5925_show_string("ON");
					break;
				case 2:
					tlc5925_show_string("AUTO");
					break;
			}
			break;
		case ST_MENU_SET_BRIGHTNESS:
			tlc5925_show_string("bright");
			break;
		case ST_SET_BRIGHTNESS:
			//tlc5925_show_number(151 - brightness*REXT/288);
			tlc5925_show_number((brightness<<2)+brightness, 0x07);	// much less code than *5
			break;
		case ST_MENU_SET_HRBEEP:
			tlc5925_show_string("hrbeep");
			break;
		case ST_SET_HRBEEP:
			if (hrbeep==0)
				tlc5925_show_string("none");
			else if (hrbeep==1)
				tlc5925_show_string("short");
			else
				tlc5925_show_string("long");
			break;
		case ST_MENU_SET_REMOTE:
			tlc5925_show_string("remote");
			break;
		case ST_MENU_REMOTE_IDENTIFY:
			tlc5925_show_string("identi");
			break;
		case ST_REMOTE_IDENTIFY:
			/*
			if (receivedPacket.protocol) {
				repeatcounter++;
				receivedPacket.protocol=0;
			}
			tlc5925_show_number(repeatcounter);
			*/
			
			switch (receivedPacket.protocol)
			{
				case 0:
					tlc5925_show_string("none");
					break;
				case 1:
					if (bits_received==20)
						tlc5925_show_string("sirc20");
					else if (bits_received==15)
						tlc5925_show_string("sirc15");
					else
						tlc5925_show_string("sirc12");
					break;
				case 2:
					tlc5925_show_string("nec");
					break;
				case 3:
					tlc5925_show_string("samsun");
					break;
				case 7:
					tlc5925_show_string("rc-5");
					break;
				case 20:
					tlc5925_show_string("jvc");
					break;
				case 28:
					tlc5925_show_string("nec42");
					break;
			}
			
			break;
		case ST_MENU_REMOTE_SHOW_CODE:
			tlc5925_show_string("shocod");
			break;
		case ST_REMOTE_SHOW_CODE:
			tlc5925_show_hex(receivedPacket.address>>8, receivedPacket.address, receivedPacket.command);
			break;
		case ST_MENU_REMOTE_LEARN:
			tlc5925_show_string("learn");
			break;
		case ST_LEARN_REMOTE_UP:
			tlc5925_show_string("up");
			break;
		case ST_LEARN_REMOTE_DOWN:
			tlc5925_show_string("down");
			break;
		case ST_LEARN_REMOTE_LEFT:
			tlc5925_show_string("left");
			break;
		case ST_LEARN_REMOTE_RIGHT:
			tlc5925_show_string("right");
			break;
		case ST_LEARN_REMOTE_BUTTON:
			tlc5925_show_string("button");
			break;
		case ST_MENU_REMOTE_RETURN:
			tlc5925_show_string("return");
			break;
		case ST_MENU_SET_VERSION:
			tlc5925_show_string("versio");
			break;
		case ST_VERSION:
			//tlc5925_show_string(FW_VERSION);
			tlc5925_show_number(FW_VERSION, 2);
			break;
		case ST_MENU_SET_RETURN:
			tlc5925_show_string("return");
			break;
	} // end display case
}

void read_jstk(void) {
	joyshadow = input_a() & 0x1f;
	
	if (joyshadow!=0x1f || joytimer)	// 15 ms debounce on press
		joytimer++;
	else
		joytimer=0;
	
	if (joyshadow==0x1f && releasetimer)			// 15 ms debounce on release
		releasetimer--;
	else if (joyshadow!=0x1f && joytimer>1)
		releasetimer=255;
	
	if (joytimer == 1 && releasetimer == 0) {
		if (!(joyshadow &0x04))
			event=EVT_DOWN;
		if (!(joyshadow &0x10))
			event=EVT_UP;
		if (!(joyshadow &0x02))
			event=EVT_LEFT;
		if (!(joyshadow &0x08))
			event=EVT_RIGHT;
		if (!(joyshadow &0x01))
			event=EVT_BUTTON;
	}		
}

void code_to_button(void) {
	event = EVT_NONE;
	if (receivedPacket.address == remoteAddress && receivedPacket.protocol == remoteProtocol) {
		if (receivedPacket.command == remoteCommandUp)
			event = EVT_UP;
		else if (receivedPacket.command == remoteCommandDown)
			event = EVT_DOWN;
		else if (receivedPacket.command == remoteCommandButton)
			event = EVT_BUTTON;
		else if (receivedPacket.command == remoteCommandLeft)
			event = EVT_LEFT;
		else if (receivedPacket.command == remoteCommandRight)
			event = EVT_RIGHT;
	}
	ir_received=FALSE;
}

void adjust_with_limits(signed int8* adjust, signed int8 low_limit, signed int8 high_limit) {
//	if (bit_test(event, 0)) {	// leads to same compiled output as the following line, but is less portable
	if (event&0x01) {	// this checks for both EVT_UP and EVT_RIGHT, but event enum must be properly arranged!
//	if (event==EVT_UP||event==EVT_RIGHT) {
		if (*adjust<high_limit)
			(*adjust)++;
		else
			*adjust=low_limit;
		break;
	}
	else {	
//	else if (event==EVT_DOWN||event==EVT_LEFT) {	// this check not needed if other events already filtered in state machine
		if (*adjust>low_limit)
			(*adjust)--;
		else
			*adjust=high_limit;
		break;
	}
}

void adjust_with_limits(int16* adjust, int16 low_limit, int16 high_limit) {
	if (event==EVT_UP) {
		if (*adjust<high_limit)
			(*adjust)++;
		else
			*adjust=low_limit;
		break;
	}
	else if (event==EVT_DOWN) {	
		if (*adjust>low_limit)
			(*adjust)--;
		else
			*adjust=high_limit;
		break;
	}
}

void save_remote() {
	write_eeprom(REMOTE_PROTOCOL_ADDR, remoteProtocol);
	write_int16_eeprom(REMOTE_ADDRESS_ADDR, remoteAddress);
	write_int16_eeprom(REMOTE_UP_ADDR, remoteCommandUp);
	write_int16_eeprom(REMOTE_DOWN_ADDR, remoteCommandDown);
	write_int16_eeprom(REMOTE_LEFT_ADDR, remoteCommandLeft);
	write_int16_eeprom(REMOTE_RIGHT_ADDR, remoteCommandRight);
	write_int16_eeprom(REMOTE_BUTTON_ADDR, remoteCommandButton);
}

void load_remote() {
	remoteProtocol = read_eeprom(REMOTE_PROTOCOL_ADDR);
	remoteAddress = read_int16_eeprom(REMOTE_ADDRESS_ADDR);
	remoteCommandUp = read_int16_eeprom(REMOTE_UP_ADDR);
	remoteCommandDown = read_int16_eeprom(REMOTE_DOWN_ADDR);
	remoteCommandLeft = read_int16_eeprom(REMOTE_LEFT_ADDR);
	remoteCommandRight = read_int16_eeprom(REMOTE_RIGHT_ADDR);
	remoteCommandButton = read_int16_eeprom(REMOTE_BUTTON_ADDR);
}

void handle_pending_tasks() {
	if (PENDING_RTC_READ) {		// RTC_READ takes about 7 ms when localtime and daylight saving time implemented
//		pcf2129_get_time_and_date(utc_time.tm_sec, utc_time.tm_min, utc_time.tm_hour, utc_time.tm_mday, utc_time.tm_wday, utc_time.tm_mon, (int8)(utc_time.tm_year));
//		utc_time.tm_mon -= 1;
//		utc_time.tm_year += 100;
		pcf2129_get_struct_tm(&utc_time);	// get struct adjusts month and year according to struct_tm spec
		if (dlight_saving<2)
			dlight_saving_on = dlight_saving;
		if (dlight_saving == 2)
			dlight_saving_on = eu_summertime(&utc_time);
		utc_to_local(&utc_time, &local_time, timezone+dlight_saving_on);
	}	
	
	if (PENDING_RTC_SAVE) {
		if ((utc_time.tm_mon == 4 || utc_time.tm_mon == 6 || utc_time.tm_mon == 9 || utc_time.tm_mon == 11)&& utc_time.tm_mday > 30)
			utc_time.tm_mday = 30;
		if (utc_time.tm_year%4 == 0 && utc_time.tm_mon == 2 && utc_time.tm_mday>29)
			utc_time.tm_mday = 29;
		else if (utc_time.tm_mon == 2 && utc_time.tm_mday>28)
			utc_time.tm_mday = 28;
		utc_to_local(&local_time, &utc_time, -(timezone+dlight_saving_on));	// use negation of timezone when converting local to utc
		pcf2129_set_time_and_date(utc_time.tm_sec, utc_time.tm_min, utc_time.tm_hour, utc_time.tm_mday, utc_time.tm_wday, utc_time.tm_mon+1, (int8)(utc_time.tm_year-100));	// this might produce less code since this function is internally used by pcf2129.c in init funcion
//		pcf2129_set_struct_tm(&utc_time); // alternative to line above, need to check which one produces more compact code
	}	
	

	if (PENDING_STATE_SAVE) {
		pcf2129_set_byte(menu_state, PCF2129_SEC_ALARM_REG);
	}	
	
	if (PENDING_BRIGHTNESS_SAVE) {
		if (RTC_init_status==2)
			write_eeprom(BRIGHTNESS_ADDR, brightness);
		else
			pcf2129_set_byte(brightness, PCF2129_MIN_ALARM_REG);
	}
	
	if (PENDING_HRBEEP_SAVE) {
		if (RTC_init_status==2)
			write_eeprom(HRBEEP_ADDR, hrbeep);
		else
			pcf2129_set_byte(hrbeep, PCF2129_HOUR_ALARM_REG);
	}
	
	if (PENDING_REMOTE_SAVE) {
		save_remote();
	}

	if (PENDING_BRIGHTNESS_SET) {
		dac_write(151 - brightness*REXT/288);
	}	

	if (PENDING_SETTINGS_SAVE) {
		write_eeprom(TIMEZONE_ADDR, timezone);
		write_eeprom(DLIGHT_SAVING_ADDR, dlight_saving);
	}	


	pending=0;
}		

void handleBuzzer() {
	if (buzzerDuration)
		setup_ccp1(CCP_PWM);
	else
		setup_ccp1(CCP_OFF);
}

void decodeGPS() {
	if (GPSDataReady) {
		if (GPS_status==2)		// First message received from GPS module, so it must be present!
			GPS_status=0;
		GPS_status = GPRMC_decode_tm(GPSData, &gps_time);
		if (GPS_status) {		// GPS data is valid (has a fix)
			utc_time = gps_time;
			if (utc_time.tm_min==59 && utc_time.tm_sec==0)	// If GPS has fix, save time&date to RTC once per hour
				PENDING_RTC_SAVE=1;
			utc_to_local(&utc_time, &local_time, timezone+dlight_saving_on);
			hourBeep();			// cases GPS_status==0 and GPS_status==2 are handled on RTC ISR
		}	
		GPSDataReady = FALSE;
	}
}

void my_write_eeprom(int8 address, int8 value) {
    write_eeprom(address,value);
}

int8 my_read_eeprom(int8 address) {
    return read_eeprom(address);
}

void show_bme_temperature() {
	if (bme280_temp<-999)	// temperatures less than -9.99 must be shown with one decimal
		tlc5925_show_number(bme280_temp,(2|0x20));
		//tlc5925_show_temperature((signed int16)(bme280_temp/10));
	else
		//tlc5925_show_number((signed int32)(bme280_temp*100),(4|0x10));
		tlc5925_show_number(bme280_temp*100,(4|0x10));
}

void hourBeep() {
	if (local_time.tm_sec>54 && local_time.tm_min==59 && hrbeep==2) // beep mode = long
		buzzerDuration=1562;	
	if (local_time.tm_sec==0 && local_time.tm_min==0 && hrbeep)
		buzzerDuration=7812;
}