////////////////////////////////////////////////////////////////////////////////
///                               TLC5925.C                                  ///
///          Driver for 16-channel constant current LED driver               ///
///                                                                          ///
/// Copyright 2017 Jaakko Kairus                                             ///
/// jaakko@kair.us                                                           ///
///                                                                          ///
/// You may freely use and modify this software for personal use.            ///
/// Commercial use of this software or parts of it is prohibited             ///
/// without written permission from author.                                  ///
/// DISCLAIMER: This software comes with ABSOLUTELY NO WARRANTY              ///
///                                                                          ///
/// For more information about this project see project web page at          ///
/// http://kair.us/projects/clock/v4/index.html                              ///
///                                                                          ///
/// Developed with CCS compiler version 5.076                                ///
///                                                                          ///
////////////////////////////////////////////////////////////////////////////////

#include "7seg_chars_kello_v4.h"
#define DIM_LEVEL 3	// defines how many dimmer function calls have some digits blanked
#define SHOW_DEGREES 0x08;
#define SHOW_RH 0x10;
int8 const seg_table_left[16] =		{_L0,_L1,_L2,_L3,_L4,_L5,_L6,_L7,_L8,_L9,	// 0,1,2,3,4,5,6,7,8,9
									 _LA,_LB,_LC,_LD,_LE,_LF}; 					// A,B,C,D,E,F
int8 const seg_table_right[16] =	{_R0,_R1,_R2,_R3,_R4,_R5,_R6,_R7,_R8,_R9,
									 _RA,_RB,_RC,_RD,_RE,_RF};
int8 const seg_table[2][16] =		{_R0,_R1,_R2,_R3,_R4,_R5,_R6,_R7,_R8,_R9,
									 _RA,_RB,_RC,_RD,_RE,_RF,
									 _L0,_L1,_L2,_L3,_L4,_L5,_L6,_L7,_L8,_L9,	// 0,1,2,3,4,5,6,7,8,9
									 _LA,_LB,_LC,_LD,_LE,_LF};				// A,B,C,D,E,F
int8 const ascii_table[2][64] =	   {_SPACE,_R_EXCLAMATION,_R_QUOTATION,_R_HASH,_R_DOLLAR,_R_PERCENT,_R_AMPERSAND,_R_APOSTROPHE,
									_R_LEFT_PARENTHESIS,_R_RIGHT_PARENTHESIS,_R_ASTERISK,_R_PLUS,_R_COMMA,_R_MINUS,_R_PERIOD,_R_SLASH,
									_R0,_R1,_R2,_R3,_R4,_R5,_R6,_R7,
									_R8,_R9,_R_COLON,_R_SEMICOLON,_R_LESS,_R_EQUALS,_R_GREATER,_R_QUESTION,
									_R_AT,_RA,_RB,_RC,_RD,_RE,_RF,_RG,
									_RH,_RI,_RJ,_RK,_RL,_RM,_RN,_RO,
									_RP,_RQ,_RR,_RS,_RT,_RU,_RV,_RW,
									_RX,_RY,_RZ,_R_LEFT_BRACKET,_R_BACKSLASH,_R_RIGHT_BRACKET,_R_CIRCUMFLEX,_R_UNDERSCORE,
									_SPACE,_L_EXCLAMATION,_L_QUOTATION,_L_HASH,_L_DOLLAR,_L_PERCENT,_L_AMPERSAND,_L_APOSTROPHE,
									_L_LEFT_PARENTHESIS,_L_RIGHT_PARENTHESIS,_L_ASTERISK,_L_PLUS,_L_COMMA,_L_MINUS,_L_PERIOD,_L_SLASH,
									_L0,_L1,_L2,_L3,_L4,_L5,_L6,_L7,
									_L8,_L9,_L_COLON,_L_SEMICOLON,_L_LESS,_L_EQUALS,_L_GREATER,_L_QUESTION,
									_L_AT,_LA,_LB,_LC,_LD,_LE,_LF,_LG,
									_LH,_LI,_LJ,_LK,_LL,_LM,_LN,_LO,
									_LP,_LQ,_LR,_LS,_LT,_LU,_LV,_LW,
									_LX,_LY,_LZ,_L_LEFT_BRACKET,_L_BACKSLASH,_L_RIGHT_BRACKET,_L_CIRCUMFLEX,_L_UNDERSCORE};
static int8 dimmer;
static int8 led_data[6];	// Display buffer

void tlc5925_show_time(int8 hours, int8 minutes, int8 seconds);
void tlc5925_show_date(int8 day, int8 month, int8 year);
void tlc5925_show_hex(int8 a, int8 b, int8 c);
void tlc5925_show_scroll(char* message,int8 speed,int8* new_press);
void tlc5925_show_string(char* message);
void tlc5925_show_time_with_dimmer(int8 hours, int8 minutes, int8 seconds, int8 dim_mode);
void tlc5925_show_date_with_dimmer(int8 day, int8 month, int8 year, int8 dim_mode);
void tlc5925_show_number(signed int32 num, int8 point_location);
void tlc5925_show_temperature(signed int16 num);
void tlc5925_send_buffer();
void get_digits(int8 value, int8* buffer);
void get_hex(int8 value, int8* buffer);

void tlc5925_show_time(int8 hours, int8 minutes, int8 seconds) {
	get_digits(seconds, &led_data[0]);
	get_digits(minutes, &led_data[2]);
	get_digits(hours, &led_data[4]);
	led_data[1] |= 0x08;
	led_data[3] |= 0x08;
	tlc5925_send_buffer();
}

void tlc5925_show_date(int8 day, int8 month, int8 year) {
	get_digits(year, &led_data[0]);
	get_digits(month, &led_data[2]);
	get_digits(day, &led_data[4]);
	led_data[2] |= 0x04;
	led_data[4] |= 0x04;
	tlc5925_send_buffer();
}

void tlc5925_show_hex(int8 a, int8 b, int8 c) {
	led_data[0]=seg_table_right[c %16];
	led_data[1]=seg_table_left[(c/16)%16];
	led_data[2]=seg_table_right[b %16];
	led_data[3]=seg_table_left[(b/16)%16];
	led_data[4]=seg_table_right[a %16];
	led_data[5]=seg_table_left[(a/16)%16];
/*
	get_hex(a, &led_data[0]);
	get_hex(b, &led_data[2]);
	get_hex(c, &led_data[4]);
*/	
	tlc5925_send_buffer();
}

//Sends scrolling string to led display via tlc5925
void tlc5925_show_scroll(char* message,int8 speed,int8* new_press)
{
	int8 i,j;
	int8 buffer[7];
	for (i=0;i<strlen(message)+8;i++) {
		if (*new_press != 0)
			return;
		for (j=0;j<6;j++) {
			if (i+j<7||i+j>strlen(message)+6) // blanks before and after message
				buffer[j]=0x20;	// ascii space
			else
				buffer[j] = message[i+j-7];
		}
		buffer[6]=0x00;	// ensure the string is null terminated
		tlc5925_show_string(buffer);	
		delay_ms(speed);
	}
}

//Sends static string to led display via tlc5925
//void tlc5925_show_string(char* message)
//{
//	int8 left,k,i=strlen(message)-1;
//	char temp;
//	for (int j=0;j<6;j++) {
//		k=i-j;
//		temp = toupper(message[k]); 
//		left=(j%2);
//		if (j>i)
//			spi_xfer(_SPACE);
//		else {  // (hopefully) a letter
//			spi_xfer(ascii_table[left][temp-' ']);
//		}
//	}	
//	output_high(TLC5925_LE);
//	output_low(TLC5925_LE);
//}
void tlc5925_show_string(char* message)
{
	int8 left,k,i=strlen(message)-1;
	char temp;
	for (int j=0;j<6;j++) {
		k=i-j;
		temp = toupper(message[k]); 
		left=(j%2);
		if (j>i)
			led_data[j]=_SPACE;
		else {  // (hopefully) a letter
			led_data[j]=ascii_table[left][temp-' '];
		}
	}	
	tlc5925_send_buffer();
}

void tlc5925_show_time_with_dimmer(int8 hours, int8 minutes, int8 seconds, int8 dim_mode) {
	if (dimmer<DIM_LEVEL) {
		dimmer++;
		memset(led_data, 0, sizeof led_data);
		if (dim_mode==3) {
			get_digits(seconds, &led_data[0]);		
		}
		if (dim_mode==2) {
			get_digits(minutes, &led_data[2]);
		}
		if (dim_mode==1) {
			get_digits(hours, &led_data[4]);
		}
		led_data[1] |= 0x08;
		led_data[3] |= 0x08;
		tlc5925_send_buffer();
	}
	else {
		dimmer=0;
		tlc5925_show_time(hours,minutes,seconds);
	}	
}

void tlc5925_show_date_with_dimmer(int8 day, int8 month, int8 year, int8 dim_mode) {
	if (dimmer<DIM_LEVEL) {
		dimmer++;
		memset(led_data, 0, sizeof led_data);
		if (dim_mode==3) {
			get_digits(year, &led_data[0]);
		}
		if (dim_mode==2) {
			get_digits(month, &led_data[2]);
		}
		if (dim_mode==1) {
			get_digits(day, &led_data[4]);
		}
		led_data[2] |= 0x04;
		led_data[4] |= 0x04;
		tlc5925_send_buffer();
	}
	else {
		dimmer=0;
		tlc5925_show_date(day,month,year);
	}	
}

void tlc5925_show_number(signed int32 num, int8 point_location) {
	int8 j,temp8;                       // Digit index
	signed int32 temp32;
	
	temp32 = labs(num); 

	for (j=0;j<6;j++) {
		temp8 = temp32 %10;
		led_data[j] = seg_table[j%2][temp8];
		temp32 /= 10;
		if (temp8 == 0 && (j>(point_location&0x07) || (((point_location&0x07)==0x07)&&j!=0)) && temp32==0)
			led_data[j] = _SPACE;	// blank leading zeros
	}

	if (num<0) {	// add minus to negative values
		for (j=0;j<6;j++) {
			if (led_data[j] == _SPACE) {
				led_data[j] = ascii_table[j%2]['-'-' '];
				j=6;
			}	
		}
	}

/*
	for (j=0;j<6;j++) {
		led_data[j] = seg_table[j%2][temp32 %10];
		temp32 /= 10;
	}

	// Blank leading zeros
	if (num<100000 && num>-10000) led_data[5] = _SPACE;
	if (num<10000 && num>-1000) led_data[4] = _SPACE;
	if (num<1000 && num>-100)  led_data[3] = _SPACE;
	if (num<100 && num>-10)   led_data[2] = _SPACE;
	if (num<10 && num>-1)    led_data[1] = _SPACE;
	
	
	// Add minuses to negative values
	if (num<-9999)	led_data[5] = _L_MINUS;
	if (num>-10000 && num<-999)	led_data[4] = _R_MINUS;
	if (num>-1000 && num<-99)	led_data[3] = _L_MINUS;
	if (num>-100 && num<-9)	led_data[2] = _R_MINUS;
	if (num>-10 && num<0)	led_data[1] = _L_MINUS;
*/
	if (point_location&0x08) {		// bit 3 of point_location shows rH
		led_data[5] = _LR;
		led_data[4] = _RH;
	}	

	if (point_location&0x10) {		// bit 4 of point_location shows degree sign and C
		led_data[1] = _L_DEGREE;
		led_data[0] = _R_CELSIUS;
	}	

	if (point_location&0x20) {		// bit 5 of point location shows just c
		led_data[0] = _RC;
	}	

	// Add decimal point if required
	point_location &= 0x07;
	if (point_location<6)
		led_data[point_location] |= _R_POINT;

   // Send led segments
	tlc5925_send_buffer();
}

void tlc5925_show_temperature(signed int16 num) {	// 8141 bytes before modding to show negs. 8244 after.
	int8 j;                       // Digit index
	signed int16 temp16;
	
	// Get led segments for all four digits. Don't need to divide to get units.
	temp16 = labs(num); 
	for (j=1;j<6;j++) {
		led_data[j] = seg_table[j%2][temp16 %10];
		temp16 /= 10;
	}

   // Do leading zero blanking
	led_data[5] = _SPACE;
	if (num<1000 && num>-100) led_data[4] = _SPACE;
	if (num<100 && num>-10)  led_data[3] = _SPACE;
	if (num<0 && labs(num)>99) led_data[4] = _R_MINUS;
	if (num<0 && labs(num)<100) led_data[3] = _L_MINUS;

	led_data[0] = _RC;
	
   // Add decimal point if required
	led_data[2] |= _R_POINT;

	tlc5925_send_buffer();	
}

void tlc5925_send_buffer() {
	for (int j=0;j<6;j++) {
		spi_xfer(led_data[j]);
	}
	output_high(TLC5925_LE);
	output_low(TLC5925_LE);	
}	
	
void get_digits(int8 value, int8* buffer) {
	*buffer = seg_table_right[value % 10];
	*(buffer+1) = seg_table_left[(value/10)%10];
}	

void get_hex(int8 value, int8* buffer) {
	*buffer = seg_table_right[value % 16];
	*(buffer+1) = seg_table_left[(value/16)%16];
}
